import logging

import opentracing
from opentracing.ext import tags as ot_tags

try:
    from flask import request
except Exception:  # pragma: no cover
    pass

from opentracing_utils.common import sanitize_url


logger = logging.getLogger(__name__)


DEFUALT_REQUEST_ATTRIBUTES = ('url', 'method')
DEFUALT_RESPONSE_ATTRIBUTES = ('status_code',)


def trace_flask(app, request_attr=DEFUALT_REQUEST_ATTRIBUTES, response_attr=DEFUALT_RESPONSE_ATTRIBUTES,
                default_tags=None, error_on_4xx=True, mask_url_query=False, mask_url_path=False, operation_name=None):
    """
    Add OpenTracing to Flask applications using ``before_request`` & ``after_request``.

    Will trace all incoming requests and add proper tags to spans.

    Will also add ``current_span`` to ``flask.request`` so other spans can access it further down the application
    functions.

    :param app: Flask application.
    :type app: Flask.App

    :param request_attr: Flask.Request attributes that can be set as span tags.
    :type request_attr: list

    :param response_attr: Flask.Response attributes that can be set as span tags.
    :type response_attr: list

    :param default_tags: Default span tags to included with every request span.
    :type default_tags: dict

    :param error_on_4xx: Set ``error`` tag in span if response is ``4xx`` or ``5xx``. Default is ``True``.
    :type error_on_4xx: bool

    :param mask_url_query: Mask URL query args in span. Default is False.
    :type mask_url_query: bool

    :param mask_url_path: Mask URL path in span. Default is False.
    :type mask_url_path: bool

    :param operation_name: Callable that returns the operation name of the request span. Default is None.
    :type operation_name: Callable[*args, **kwargs]
    """

    min_error_code = 400 if error_on_4xx else 500

    @app.before_request
    def trace_request():
        op_name = request.endpoint if request.endpoint else request.path.strip('/').replace('/', '_')

        if callable(operation_name):
            op_name = operation_name()

        span = None
        headers_carrier = dict(request.headers.items())

        try:
            span_ctx = opentracing.tracer.extract(opentracing.Format.HTTP_HEADERS, headers_carrier)
            span = opentracing.tracer.start_span(operation_name=op_name, child_of=span_ctx)
        except (opentracing.InvalidCarrierException, opentracing.SpanContextCorruptedException):
            span = opentracing.tracer.start_span(operation_name=op_name, tags={'flask-no-propagation': True})

        if span is None:
            span = opentracing.tracer.start_span(op_name)

        if request_attr:
            for attr in request_attr:
                if hasattr(request, attr):
                    try:
                        tag_value = str(getattr(request, attr))
                        tag_key = attr

                        # Masking URL query and path params.
                        if attr == 'url':
                            tag_value = sanitize_url(
                                tag_value, mask_url_query=mask_url_query, mask_url_path=mask_url_path)
                            tag_key = ot_tags.HTTP_URL
                        elif attr == 'method':
                            tag_key = ot_tags.HTTP_METHOD

                        if tag_value:
                            span.set_tag(tag_key, tag_value)
                    except Exception:
                        pass

        if type(default_tags) is dict:
            for k, v in default_tags.items():
                try:
                    span.set_tag(k, v)
                except Exception:
                    pass

        span.set_tag(ot_tags.COMPONENT, 'flask')
        span.set_tag(ot_tags.SPAN_KIND, ot_tags.SPAN_KIND_RPC_SERVER)

        # Use ``flask.request`` as in process context.
        request.current_span = span

    @app.after_request
    def trace_response(response):
        try:
            if hasattr(request, 'current_span'):
                if response_attr:
                    for attr in response_attr:
                        if hasattr(response, attr):
                            tag_key = attr
                            if tag_key == 'status_code':
                                tag_key = ot_tags.HTTP_STATUS_CODE
                            request.current_span.set_tag(tag_key, str(getattr(response, attr)))

                if response.status_code >= min_error_code:
                    request.current_span.set_tag('error', True)

                request.current_span.finish()
        finally:
            return response


def extract_span_from_flask_request(*args, **kwargs):
    """
    Safe utility function to extract the ``current_span`` from ``flask.request``. Compatible with ``@trace`` decorator.
    """
    try:
        return getattr(request, 'current_span', None)
    except Exception:  # pragma: no cover
        pass

    return None



import os
import logging

import opentracing

OPENTRACING_INSTANA = 'instana'
OPENTRACING_LIGHTSTEP = 'lightstep'
OPENTRACING_JAEGER = 'jaeger'
OPENTRACING_BASIC = 'basic'


logger = logging.getLogger(__name__)


def init_opentracing_tracer(tracer, **kwargs):
    if tracer == OPENTRACING_BASIC:
        from basictracer import BasicTracer  # noqa

        recorder = kwargs.get('recorder')
        sampler = kwargs.get('sampler')

        opentracing.tracer = BasicTracer(recorder=recorder, sampler=sampler)
    elif tracer == OPENTRACING_INSTANA:
        import instana.options as InstanaOpts
        import instana.tracer  # noqa

        service = kwargs.pop(
            'service',
            os.environ.get('OPENTRACING_INSTANA_SERVICE'))
        log_level = kwargs.pop('log_level', logging.INFO)

        instana.tracer.init(InstanaOpts.Options(service=service, log_level=log_level, **kwargs))
    elif tracer == OPENTRACING_LIGHTSTEP:
        import lightstep

        # Get Lightstep specific config vars
        component_name = kwargs.pop(
            'component_name',
            os.environ.get('OPENTRACING_LIGHTSTEP_COMPONENT_NAME'))
        access_token = kwargs.pop(
            'access_token',
            os.environ.get('OPENTRACING_LIGHTSTEP_ACCESS_TOKEN'))
        collector_host = kwargs.pop(
            'collector_host',
            os.environ.get('OPENTRACING_LIGHTSTEP_COLLECTOR_HOST', 'collector.lightstep.com'))
        collector_port = kwargs.pop(
            'collector_port',
            int(os.environ.get('OPENTRACING_LIGHTSTEP_COLLECTOR_PORT', 443)))
        verbosity = kwargs.pop(
            'verbosity',
            int(os.environ.get('OPENTRACING_LIGHTSTEP_VERBOSITY', 0)))

        if not access_token:
            logger.warn('Initializing LighStep tracer with no access_token!')

        opentracing.tracer = lightstep.Tracer(
            component_name=component_name, access_token=access_token, collector_host=collector_host,
            collector_port=collector_port, verbosity=verbosity, **kwargs)
    elif tracer == OPENTRACING_JAEGER:
        from jaeger_client import Config

        service_name = kwargs.pop('service_name', os.environ.get('OPENTRACING_JAEGER_SERVICE_NAME'))
        config = kwargs.pop('config', {})

        jaeger_config = Config(config=config, service_name=service_name)
        opentracing.tracer = jaeger_config.initialize_tracer()
    else:
        opentracing.tracer = opentracing.Tracer()



import functools

from opentracing_utils.span import get_new_span, adjust_span, get_span_from_kwargs, remove_span_from_kwargs


def trace(component=None, operation_name=None, tags=None, use_follows_from=False, pass_span=False, inspect_stack=True,
          ignore_parent_span=False, span_extractor=None, skip_span=None):
    """
    Opentracing tracer decorator. Attempts to extract parent span and create a new span for the decorated function.

    :param commponent: commponent name.
    :type commponent: str

    :param operation: Span operation name. If not set then it will be the ``function.__name__``.
    :type operation: str

    :param tags: Span tags. If tags is not a dict, then it will be ignored.
    :type tags: dict

    :param use_follows_from: Use a ``follows_from`` reference instead of the default ``child_of``.
    :type use_follows_from: bool

    :param pass_span: Whether to pass the newly created span to the decorated function. Default is False.
                      This is useful if the traced function would add extra tags to the span.
                      *IMPORTANT*: This will not pass the parent span if any, it will only pass the new span decorating
                      the traced function even if the caller passed the parent span explicitly.
    :type pass_span: bool

    :param inspect_stack: Whether to inspect call stack frames to retrieve the parent span. Default is True.
    :type inspect_stack: bool

    :param ignore_parent_span: Always start a fresh span for this decorated function. Default is False.
    :type ignore_parent_span: bool

    :param span_extractor: A callable that retrieves the current parent span. This is useful when decorator is used
                           in a context aware application and there is an alternative way to retrieve the span.
                           Callable should expect the traced function *args & **kwargs.
                           If None is returned from the callable, then normal span extraction will be applied.
    :type span_extractor: Callable[*args, **kwargs]

    :param skip_span: A callable which can be used to determine if the span should be skipped.
    :type skip_span: Callable[*args, **kwargs]
    """

    def trace_decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if skip_span is not None and skip_span(*args, **kwargs):
                return f(*args, **remove_span_from_kwargs(**kwargs))

            # Get a new current span wrapping this traced function.
            # ``get_new_span`` should retrieve parent_span if any!
            span_arg_name, current_span = get_new_span(
                f, args, kwargs, inspect_stack=inspect_stack, ignore_parent_span=ignore_parent_span,
                span_extractor=span_extractor, use_follows_from=use_follows_from)

            if pass_span and span_arg_name:
                kwargs[span_arg_name] = current_span
            else:
                if span_extractor:
                    kwarg_span, _ = get_span_from_kwargs(**kwargs)
                    kwargs.pop(kwarg_span, None)

                kwargs.pop(span_arg_name, None)

            current_span = adjust_span(current_span, operation_name, component, tags)

            with current_span:
                return f(*args, **kwargs)

        return wrapper

    return trace_decorator



import opentracing

try:
    from sqlalchemy.engine import Engine
    from sqlalchemy.event import listens_for
except ImportError:  # pragma: no cover
    pass

from opentracing.ext import tags as ot_tags
from opentracing_utils.span import get_parent_span


def trace_sqlalchemy(operation_name=None, span_extractor=None, set_error_tag=False):
    """
    Trace Sqlalchemy database queries.

    :param operation_name: Callable to return the operation name of the query. By default, operation_name will be the
                           clause of the SQL statement (e.g. select, update, delete).
    :type operation_name: Callable[conn, cursor, statement, parameters, context, executemany]

    :param span_extractor: Callable to return the parent span. Default is ``None``, and trace_sqlachemy will attempt
                           to detect the parent span.
    :type span_extractor: Callable[conn, cursor, statement, parameters, context, executemany]

    :param set_error_tag: Database query span will set error tag in case of any exceptions. Default is False.
    :type set_error_tag: bool
    """

    @listens_for(Engine, 'before_cursor_execute')
    def trace_before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
        if callable(span_extractor):
            parent_span = span_extractor(conn, cursor, statement, parameters, context, executemany)
        else:
            _, parent_span = get_parent_span()

        if context:
            op_name = statement.split(' ')[0].lower() or 'query'
            if callable(operation_name):
                op_name = operation_name(conn, cursor, statement, parameters, context, executemany)

            query_span = opentracing.tracer.start_span(operation_name=op_name, child_of=parent_span)

            if context:
                (query_span
                    .set_tag(ot_tags.COMPONENT, 'sqlalchemy')
                    .set_tag('db.type', 'sql')
                    .set_tag('db.engine', context.dialect.name)
                    .set_tag('db.statement', statement))

                context._query_span = query_span

    @listens_for(Engine, 'after_cursor_execute')
    def tarce_after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
        if hasattr(context, '_query_span'):
            context._query_span.finish()

    @listens_for(Engine, 'handle_error')
    def trace_handle_error(exception_context):
        context = exception_context.execution_context
        if hasattr(context, '_query_span'):
            context._query_span.log_kv({'exception': str(exception_context.original_exception)})

            if set_error_tag:
                context._query_span.set_tag('error', True)

            context._query_span.finish()



from pkg_resources import get_distribution

from opentracing_utils.decorators import trace

from opentracing_utils.tracers import init_opentracing_tracer
from opentracing_utils.tracers import OPENTRACING_BASIC, OPENTRACING_INSTANA, OPENTRACING_LIGHTSTEP, OPENTRACING_JAEGER

from opentracing_utils.span import extract_span_from_kwargs, remove_span_from_kwargs

from opentracing_utils.libs._requests import trace_requests, sanitize_url
from opentracing_utils.libs._flask import trace_flask, extract_span_from_flask_request
from opentracing_utils.libs._sqlalchemy import trace_sqlalchemy


__version__ = get_distribution('opentracing-utils').version


__all__ = (
    'extract_span_from_flask_request',
    'extract_span_from_kwargs',
    'init_opentracing_tracer',
    'remove_span_from_kwargs',
    'sanitize_url',
    'trace',
    'trace_flask',
    'trace_requests',
    'trace_sqlalchemy',

    'OPENTRACING_BASIC',
    'OPENTRACING_INSTANA',
    'OPENTRACING_JAEGER',
    'OPENTRACING_LIGHTSTEP',
)



from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa

import urllib.parse as parse


def sanitize_url(url, mask_url_query=True, mask_url_path=False):
    parsed = parse.urlsplit(url)

    # masking - may be give some hints in masking query and path instead of '?' ??
    host = '{}:{}'.format(parsed.hostname, parsed.port) if parsed.port else parsed.hostname
    query = str(parse.urlencode({k: '?' for k in parse.parse_qs(parsed.query).keys()})) if \
        mask_url_query else parsed.query
    path = '/??/' if parsed.path and mask_url_path else parsed.path

    components = parse.SplitResult(parsed.scheme, host, path, query, parsed.fragment)

    return parse.urlunsplit(components)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Setup file for opentracing-utils.
"""
import io
import sys
from setuptools import setup, find_packages

PY3 = sys.version_info.major == 3

MAIN_PACKAGE = 'opentracing_utils'
VERSION = '0.14'
DESCRIPTION = 'OpenTracing utilities library'


def load_req(req):
    try:
        return [r.strip() for r in open(req).read().splitlines() if r.strip() and not r.strip().startswith('#')]
    except Exception:
        return []


TEST_REQUIREMENTS = load_req('test_requirements.txt')


if __name__ == '__main__':
    setup(
        name='opentracing-utils',
        version=VERSION,
        description=DESCRIPTION,
        long_description=io.open('README.rst', encoding='utf-8').read(),
        license='The MIT License (MIT)',
        url='https://github.com/zalando-zmon/opentracing-utils',
        packages=find_packages(exclude=['tests']),
        install_requires=[
            'future',
            'opentracing',
        ],
        setup_requires=[
            'opentracing',
            'pytest-runner',
        ],
        test_suite='tests',
        tests_require=TEST_REQUIREMENTS,
        include_package_data=True,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python',
            'Programming Language :: Python :: Implementation :: CPython',
            'Environment :: Console',
            'Operating System :: POSIX :: Linux',
            'Operating System :: MacOS :: MacOS X',
            'Topic :: Utilities',
            'Topic :: System :: Monitoring',
            'Topic :: System :: Networking :: Monitoring',
        ]
    )
